'''Contains the Messages data structure.'''
from typing import (Dict, List, Mapping)

_GENERAL_SUBJECT = ''

MessageData = Mapping[str, List[str]]

class Messages:
  '''Contains a mapping of subjects to a list of messages.'''
  def __init__(self) -> None:
    self._data: Dict[str, List[str]] = {}

  def __bool__(self):
    return self.has_messages

  @property
  def has_messages(self) -> bool:
    return bool(self._data)

  @property
  def messages(self) -> MessageData:
    return self._data

  def add_message(self, subject: str, message: str) -> 'Messages':
    if subject not in self._data:
      self._data[subject] = []
    self._data[subject].append(message)
    return self

  def add_general_message(self, message: str) -> 'Messages':
    return self.add_message(_GENERAL_SUBJECT, message)

  def has_message_for(self, subject: str) -> bool:
    return subject in self._data

  def has_general_message(self) -> bool:
    return self.has_message_for(_GENERAL_SUBJECT)



'''Dict serialization methods.'''
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    TypeVar,
)
from base64 import (b64encode)
from datetime import (date as Date, datetime as DateTime)
from decimal import (Decimal)
from uuid import (UUID)

T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')
Record = Mapping[str, Any]
Bytes = bytes

class Serialize:
  '''Contains serialization methods. They're placed as static methods on a class to avoid
  name conflicts.
  '''
  @staticmethod
  def map(fn: Callable[[T], U]) -> Callable[[Iterable[T]], List[U]]:
    '''Returns a function that accepts a iterable and maps over it with fn.'''
    def mapper(items: Iterable[T]) -> List[U]:
      return [fn(item) for item in items]
    return mapper

  @staticmethod
  def pick(keys: Sequence[str]) -> Callable[[Record], Record]:
    '''Returns a new mapping with only the keys specified.'''
    def picker(item: Record) -> Record:
      return {k: item[k] for k in keys if k in item}
    return picker

  @staticmethod
  def exclude(keys: Set[str]) -> Callable[[Record], Record]:
    '''Returns a new mapping without the keys specified.'''
    def excluder(item: Record) -> Record:
      return {k: v for (k, v) in item.items() if k not in keys}
    return excluder

  @staticmethod
  def evolve(maps: Mapping[str, Callable[[Any], Any]]) -> Callable[[Record], Record]:
    '''Returns a new mapping with certain keys mapped as specified.'''
    def evolver(item: Record) -> Record:
      return {k: (maps[k](v) if k in maps else v) for (k, v) in item.items()}
    return evolver

  @staticmethod
  def compose(outer: Callable[[U], V], inner: Callable[[T], U]) -> Callable[[T], V]:
    '''Returns a single function that composes outer of inner.'''
    def composer(val: T) -> V:
      return outer(inner(val))
    return composer

  @staticmethod
  def optional(fn: Callable[[T], U]) -> Callable[[Optional[T]], Optional[U]]:
    '''Returns a function that will return None if given None, otherwise maps the value with fn.'''
    def optionaler(val: Optional[T]) -> Optional[U]:
      if val is None:
        return None
      return fn(val)
    return optionaler

  @staticmethod
  def date(val: Date) -> str:
    '''date to ISO format.'''
    return val.isoformat()

  @staticmethod
  def datetime(val: DateTime) -> str:
    '''datetime to ISO format.'''
    return val.isoformat()

  @staticmethod
  def uuid(val: UUID) -> str:
    '''UUID to string.'''
    return str(val)

  @staticmethod
  def decimal(val: Decimal) -> str:
    '''Decimal number to string.'''
    return str(val)

  @staticmethod
  def bytes(val: Bytes) -> str:
    '''Bytes to b64 string.'''
    return b64encode(val).decode('utf-8')



'''Validator and Validations.'''
import re
import inspect
from base64 import (b64decode)
from dataclasses import (dataclass)
from uuid import (UUID)
from typing import (
    cast,
    Any,
    Awaitable,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
)
from .validation_result import (ValidationResult)

EMPTY_UUID = UUID(fields=(0, 0, 0, 0, 0, 0))
EMPTY_BYTES = b''

Record = Mapping[str, Any]
MutableRecord = MutableMapping[str, Any]
TryParseResult = Tuple[bool, Any]
ValidateFieldSync = Callable[[Any], bool]
ValidateFieldAsync = Callable[[Any], Awaitable[bool]]
ValidateField = Union[ValidateFieldSync, ValidateFieldAsync]
ValidateRecordSync = Callable[[Record], bool]
ValidateRecordAsync = Callable[[Record], Awaitable[bool]]
ValidateRecord = Union[ValidateRecordSync, ValidateRecordAsync]
ValidateTryParseSync = Callable[[Any], TryParseResult]
ValidateTryParseAsync = Callable[[Any], Awaitable[TryParseResult]]
ValidateTryParse = Union[ValidateTryParseSync, ValidateTryParseAsync]

@dataclass
class FieldValidator:
  '''Information to validate a single field.'''
  field_name: str
  message: str
  validator: ValidateField

  async def validate(self, value: Any) -> bool:
    '''Properly calls the validator.'''
    if inspect.iscoroutinefunction(self.validator):
      return await cast(ValidateFieldAsync, self.validator)(value)
    return cast(ValidateFieldSync, self.validator)(value)

@dataclass
class RecordValidator:
  '''Information to validate a whole record.'''
  message: str
  validator: ValidateRecord
  field_name: Optional[str]

  async def validate(self, record: Record) -> bool:
    '''Properly calls the validator.'''
    if inspect.iscoroutinefunction(self.validator):
      return await cast(ValidateRecordAsync, self.validator)(record)
    return cast(ValidateRecordSync, self.validator)(record)

@dataclass
class TryParseValidator:
  '''Information to try to parse a single field.'''
  field_name: str
  message: str
  validator: ValidateTryParse

  async def validate(self, value: Any) -> TryParseResult:
    '''Properly calls the validator.'''
    if inspect.iscoroutinefunction(self.validator):
      return await cast(ValidateTryParseAsync, self.validator)(value)
    return cast(ValidateTryParseSync, self.validator)(value)

ValidatorType = Union[FieldValidator, RecordValidator, TryParseValidator]

class Validator:
  '''Records validators for a record.'''
  def __init__(self, validators: Optional[List[ValidatorType]] = None):
    self.validators: List[ValidatorType] = validators or []

  @staticmethod
  def field(field_name, message: str, validate: ValidateField) -> FieldValidator:
    return FieldValidator(field_name, message, validate)

  @staticmethod
  def record(message: str, validate: ValidateRecord, field_name: str = None) -> RecordValidator:
    return RecordValidator(message, validate, field_name)

  @staticmethod
  def try_parse(
      field_name: str,
      message: str,
      validate: ValidateTryParse,
  ) -> TryParseValidator:
    '''Builds a TryParseValidator'''
    return TryParseValidator(field_name, message, validate)

  async def validate(self, record: Record) -> ValidationResult[MutableRecord]:
    return await ValidatorInstance(self, record).validate()

class ValidatorInstance:
  '''Holds results for a run of validation.'''
  def __init__(self, validator: Validator, record: Record) -> None:
    self.validator = validator
    self.record = record
    self.output: MutableRecord = {}
    self.validation_result = ValidationResult[MutableRecord](None)

  async def validate(self) -> ValidationResult[MutableRecord]:
    '''Executes all validators against the record until one fails.'''
    for validator in self.validator.validators:
      if isinstance(validator, FieldValidator):
        await self.validate_field(validator)
      elif isinstance(validator, RecordValidator):
        await self.validate_record(validator)
      elif isinstance(validator, TryParseValidator):
        await self.try_parse_field(validator)
      else:
        raise Exception('Invalid validator')

      if not self.validation_result:
        return self.validation_result

    record = self.record or {}
    output = {**record, **self.output}
    self.validation_result.set_output(output)
    return self.validation_result

  async def validate_field(self, validator: FieldValidator) -> None:
    '''Validate a single field.'''
    field_name = validator.field_name

    if field_name not in self.output:
      value = self.record.get(field_name, None)
      self.output[field_name] = value
    else:
      value = self.output[field_name]

    result = await validator.validate(value)

    if not result:
      self.validation_result.add_message(field_name, validator.message)

  async def validate_record(self, validator: RecordValidator) -> None:
    '''Validate an entire record.'''
    record: Record

    if self.record:
      record = {**self.record, **self.output}
    else:
      record = self.output

    result = await validator.validate(record)

    if not result:
      if validator.field_name:
        self.validation_result.add_message(validator.field_name, validator.message)
      else:
        self.validation_result.add_general_message(validator.message)

  async def try_parse_field(self, validator: TryParseValidator) -> None:
    '''Try to parse a field.'''
    field_name = validator.field_name

    if field_name not in self.output:
      value = self.record.get(field_name, None)
      self.output[field_name] = value
    else:
      value = self.output[field_name]

    (is_ok, output) = await validator.validate(value)

    if is_ok:
      self.output[field_name] = output
    else:
      self.validation_result.add_message(field_name, validator.message)

class Validators:
  '''Contains validator methods.'''
  @staticmethod
  def is_present(val: Any) -> bool:
    '''Makes sure a value is present.'''
    if val is None:
      return False
    if val == 0:
      return True
    return bool(val)

  @staticmethod
  def is_match(pattern: str) -> Callable[[str], bool]:
    '''Makes sure a string matches a value.'''
    def matcher(val: str) -> bool:
      if val is None:
        return False
      return bool(re.match(pattern, val))
    return matcher

  @staticmethod
  def optional(validator: Callable[[Any], Any]) -> Callable[[Any], Any]:
    '''Wraps a validator in an optional flag allowing None values.'''
    def checker(val: Any) -> bool:
      if val is None:
        return True
      return validator(val)
    return checker

  @staticmethod
  def try_parse_optional(
      validator: Callable[[Any], TryParseResult],
  ) -> Callable[[Any], TryParseResult]:
    '''Wraps a validator in an optional flag allowing None values.'''
    def checker(val: Any) -> TryParseResult:
      if val is None:
        return (True, val)
      return validator(val)
    return checker

  @staticmethod
  def try_parse_int(val: Any) -> Tuple[bool, int]:
    '''Tries to parse an int.'''
    if val is None:
      return (False, 0)
    try:
      return (True, int(val))
    except ValueError:
      return (False, 0)

  @staticmethod
  def try_parse_float(val: Any) -> Tuple[bool, float]:
    '''Tries to parse a float.'''
    if val is None:
      return (False, 0.0)
    try:
      return (True, float(val))
    except ValueError:
      return (False, 0.0)

  @staticmethod
  def try_parse_base64(val: Any) -> Tuple[bool, bytes]:
    '''Tries to parse a base64 string to bytes.'''
    if val is None:
      return (False, EMPTY_BYTES)
    try:
      return (True, b64decode(val, validate=True))
    except ValueError:
      return (False, EMPTY_BYTES)

  @staticmethod
  def try_parse_uuid(val: Any) -> Tuple[bool, UUID]:
    '''Tries to parse a UUID.'''
    if val is None:
      return (False, EMPTY_UUID)
    if isinstance(val, UUID):
      return (True, val)
    try:
      return (True, UUID(val))
    except (ValueError, TypeError):
      return (False, EMPTY_UUID)

  @staticmethod
  def is_length(min_len: Optional[int], max_len: Optional[int]) -> Callable[[str], bool]:
    '''Checks if a string is in length bounds.'''
    def checker(val: str) -> bool:
      if min_len is not None and len(val) < min_len:
        return False
      if max_len is not None and max_len < len(val):
        return False
      return True
    return checker



'''Contains the TransactionBase class.'''
from typing import (Optional, Tuple)
from ntier.transaction_result import TransactionResult
from ntier.constants import (
    PAGE_DEFAULT,
    PER_PAGE_DEFAULT,
    PER_PAGE_MAX,
)

class TransactionBase:
  '''Base class for transactions.'''
  def __init__(self):
    self.paging: Optional[Tuple[int, int]] = None

  def set_paging(self, page: int, per_page: int) -> 'TransactionBase':
    '''Sets paging values for the transaction.'''
    if page < 1:
      page = 1
    if per_page < 1:
      per_page = 1
    if per_page > PER_PAGE_MAX:
      per_page = PER_PAGE_MAX

    self.paging = (page, per_page)
    return self

  @property
  def offset(self):
    '''Returns the offset based on the paging.'''
    if self.paging is None:
      (page, per_page) = (PAGE_DEFAULT, PER_PAGE_DEFAULT)
    else:
      (page, per_page) = self.paging
    return (page - 1) * per_page

  @property
  def limit(self):
    '''Returns the limit based on the paging.'''
    if self.paging is None:
      return PER_PAGE_DEFAULT
    (_, per_page) = self.paging
    return per_page

  def set_output_paging(self, result: TransactionResult, total_records: int) -> TransactionResult:
    '''Assign paging to a TransactionResult.'''
    if self.paging is None:
      (page, per_page) = (PAGE_DEFAULT, PER_PAGE_DEFAULT)
    else:
      (page, per_page) = self.paging
    result.set_paging(page, per_page, total_records)
    return result



'''Contains the TransactionCode Enum.'''
from enum import Enum

class TransactionCode(Enum):
  '''This enum represents all the status codes that can be returned in a TransactionResult.'''
  success = 100
  found = 101
  created = 102
  updated = 103
  not_changed = 104
  deleted = 105
  failed = 200
  not_found = 201
  not_authenticated = 202
  not_authorized = 203
  not_valid = 204



'''Contains the transaction Data class.'''
from dataclasses import (dataclass)
from typing import (Any, Generic, MutableMapping, TypeVar)

TransactionState = MutableMapping[str, Any]
InputData = TypeVar('InputData')

@dataclass(frozen=True)
class TransactionData(Generic[InputData]):
  '''Contains input and state data for an API Transaction.'''
  input: InputData
  state: TransactionState



'''Contains the PolicyResult data structure.'''
from typing import (Optional)
from .messages import (MessageData, Messages)

class PolicyResult:
  '''Contains information about whether a policy passes or fails.'''
  def __init__(self, is_valid: bool, messages: Optional[Messages]) -> None:
    self.is_valid: bool = is_valid
    self._messages: Optional[Messages] = messages

  def __bool__(self) -> bool:
    return self.is_valid

  @property
  def messages(self) -> MessageData:
    if self._messages is None:
      return {}
    return self._messages.messages

  @classmethod
  def success(cls) -> 'PolicyResult':
    return PolicyResult(True, None)

  @classmethod
  def failed(cls, message: str) -> 'PolicyResult':
    return PolicyResult(False, Messages().add_general_message(message))



'''Contains the Paging data structure.'''
import math
from dataclasses import dataclass

@dataclass(frozen=True)
class Paging:
  '''Contains paging information.'''
  page: int
  per_page: int
  total_records: int
  total_pages: int

  @classmethod
  def build(cls, page: int, per_page: int, total_records: int) -> 'Paging':
    return cls(page, per_page, total_records, math.ceil(total_records / per_page))



'''Contains the TransactionResult data structure.'''
from typing import (Any, Optional)
from ntier.messages import (Messages)
from ntier.paging import (Paging)
from ntier.policy_result import (PolicyResult)
from ntier.transaction_code import (TransactionCode)
from ntier.validation_result import (ValidationResult)

class TransactionResult:
  '''Represents a standard transaction result.'''
  def __init__(self, is_valid: bool, status_code: TransactionCode, payload: Any) -> None:
    self.is_valid: bool = is_valid
    self.status_code: TransactionCode = status_code
    self.payload: Any = payload
    self._paging: Optional[Paging] = None

  def __repr__(self):
    return '{}(is_valid={}, status_code={}, payload={}, paging={})'.format(
        self.__class__.__name__,
        self.is_valid,
        self.status_code,
        self.payload,
        self._paging,
    )

  @property
  def paging(self) -> Paging:
    if self._paging is None:
      raise Exception('Paging is not set.')
    return self._paging

  @property
  def has_paging(self) -> bool:
    return self._paging is not None

  def set_paging(self, page: int, per_page: int, total_records: int) -> 'TransactionResult':
    self._paging = Paging.build(page, per_page, total_records)
    return self

  @classmethod
  def success(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.success, payload)

  @classmethod
  def found(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.found, payload)

  @classmethod
  def created(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.created, payload)

  @classmethod
  def updated(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.updated, payload)

  @classmethod
  def not_changed(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.not_changed, payload)

  @classmethod
  def deleted(cls, payload: Any = None) -> 'TransactionResult':
    return cls(True, TransactionCode.deleted, payload)

  @classmethod
  def failed(cls, payload: Any = None) -> 'TransactionResult':
    return cls(False, TransactionCode.failed, payload)

  @classmethod
  def not_found(cls, subject: str) -> 'TransactionResult':
    msg = Messages().add_general_message(f'{subject} not found')
    return cls(False, TransactionCode.not_found, msg.messages)

  @classmethod
  def not_authenticated(cls) -> 'TransactionResult':
    return cls(False, TransactionCode.not_authenticated, None)

  @classmethod
  def not_authorized(cls, policy_result: PolicyResult) -> 'TransactionResult':
    return cls(False, TransactionCode.not_authorized, policy_result.messages)

  @classmethod
  def not_valid(cls, validation_result: ValidationResult) -> 'TransactionResult':
    return cls(False, TransactionCode.not_valid, validation_result.messages)



'''Contains the ValidationResult data structure.'''
from typing import (Generic, Optional, TypeVar)
from .messages import (MessageData, Messages)

T = TypeVar('T')

class ValidationResult(Generic[T]):
  '''Contains information about whether a validation passes or fails.'''
  def __init__(self, messages: Optional[Messages] = None) -> None:
    self._messages: Optional[Messages] = messages
    self.output: Optional[T] = None

  def __bool__(self) -> bool:
    return self.is_valid

  @property
  def is_valid(self):
    return not bool(self.messages)

  @property
  def messages(self) -> MessageData:
    if self._messages is None:
      return {}
    return self._messages.messages

  def set_output(self, output: T) -> 'ValidationResult':
    self.output = output
    return self

  def add_message(self, subject: str, message: str) -> 'ValidationResult':
    if self._messages is None:
      self._messages = Messages()
    self._messages.add_message(subject, message)
    return self

  def add_general_message(self, message: str) -> 'ValidationResult':
    if self._messages is None:
      self._messages = Messages()
    self._messages.add_general_message(message)
    return self

  @classmethod
  def success(cls) -> 'ValidationResult':
    return ValidationResult(None)

  @classmethod
  def failed(cls, messages: Messages) -> 'ValidationResult':
    return ValidationResult(messages)

  @classmethod
  def builder(cls) -> 'ValidationResult':
    return ValidationResult(Messages())



'''ntier API.'''
from ntier.api_transaction_base import APITransactionBase
from ntier.data import (TransactionData, TransactionState)
from ntier.messages import Messages
from ntier.paging import Paging
from ntier.policy_result import PolicyResult
from ntier.transaction_base import TransactionBase
from ntier.transaction_code import TransactionCode
from ntier.transaction_result import TransactionResult
from ntier.validation_result import ValidationResult
from ntier.validator import (Validator, Validators)



'''Contains the StructuredTransaction class.'''
# pylint: disable=unused-argument
from typing import (Generic, Optional)
from ntier.data import (InputData, TransactionData, TransactionState)
from ntier.policy_result import PolicyResult
from ntier.transaction_base import TransactionBase
from ntier.transaction_result import TransactionResult
from ntier.validation_result import ValidationResult

class APITransactionBase(TransactionBase, Generic[InputData]):
  '''Allows child classes to define parts of a standard API request life-cycle:
    - initialize
    - authenticate
    - find
    - authorize
    - validate
    - perform
  '''
  async def execute(self, input_data: InputData) -> TransactionResult:
    '''Calls and response properly to overridden methods to run a standard API request.'''
    state = await self.initialize(input_data)
    data: TransactionData[InputData] = TransactionData(input_data, state)

    is_authenticated = await self.authenticate(data)
    if not is_authenticated:
      return TransactionResult.not_authenticated()

    missing_entity_name = await self.find(data)
    if missing_entity_name:
      return TransactionResult.not_found(missing_entity_name)

    policy_result = await self.authorize(data)
    if not policy_result.is_valid:
      return TransactionResult.not_authorized(policy_result)

    validation_result = await self.validate(data)
    if not validation_result.is_valid:
      return TransactionResult.not_valid(validation_result)

    result = await self.perform(data)
    return result

  async def __call__(self, input_data: InputData) -> TransactionResult:
    return await self.execute(input_data)

  async def initialize(self, input_data: InputData) -> TransactionState:
    return {}

  async def authenticate(self, data: TransactionData[InputData]) -> bool:
    return True

  async def find(self, data: TransactionData[InputData]) -> Optional[str]:
    return None

  async def authorize(self, data: TransactionData[InputData]) -> PolicyResult:
    return PolicyResult.success()

  async def validate(self, data: TransactionData[InputData]) -> ValidationResult:
    return ValidationResult.success()

  async def perform(self, data: TransactionData[InputData]) -> TransactionResult:
    return TransactionResult.success()
